@@ -1822,6 +1822,8 @@ of buffer-file-coding-system set by this function."
     ("sp" . "Latin-5") ; Serbian (Cyrillic alphabet), e.g. X11R6.4
     ("su" . "Latin-1") ; Finnish, e.g. Solaris 2.6
     ("jp" . "Japanese") ; e.g. MS Windows
+    ("chs" . "Chinese-GB") ; MS Windows Chinese Simplified
+    ("cht" . "Chinese-BIG5") ; MS Windows Chinese Traditional
     )
   "List of pairs of locale regexps and language names.
 The first element whose locale regexp matches the start of a downcased locale
@@ -1849,6 +1851,7 @@ the language name that would otherwise be used for this locale.")
     ("ja.*[._]jis7" . iso-2022-jp)
     ("ja.*[._]pck" . japanese-shift-jis)
     ("ja.*[._]sjis" . japanese-shift-jis)
+    ("jpn" . japanese-shift-jis)   ; MS-Windows uses this.
     )
   "List of pairs of locale regexps and preferred coding systems.
 The first element whose locale regexp matches the start of a downcased locale
